/*
 * $XFree86: $
 *
 * Copyright  2002 Carl D. Worth
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of Carl
 * D. Worth not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  Carl D. Worth makes no representations about the
 * suitability of this software for any purpose.  It is provided "as
 * is" without express or implied warranty.
 *
 * CARL D. WORTH DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL CARL D. WORTH BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Carl Worth, USC/Information Sciences Institute */

#include <stdlib.h>

#include "xcint.h"

static const XTransform XC_XTRANSFORM_IDENTITY = {
    {
	{65536,     0,     0},
	{    0, 65536,     0},
	{    0,     0, 65536}
    }
};

static char *
_XcRenderFilterName (XcFilter filter);

XcSurface *
XcSurfaceCreateForDrawable (Display	*dpy,
			    Drawable	drawable, 
			    Visual	*visual,
			    int		render_format,
			    Colormap	colormap)
{
    XcSurface *surface;

    surface = malloc (sizeof (XcSurface));
    if (surface == NULL)
	return NULL;

    surface->type = XC_SURFACE_TYPE_DRAWABLE;
    surface->transform = XC_XTRANSFORM_IDENTITY;

    surface->dpy = dpy;
    surface->gc = 0;
    surface->drawable = drawable;
    surface->visual = visual;

    if (! XRenderQueryVersion (dpy, &surface->render_major, &surface->render_minor)) {
	surface->render_major = -1;
	surface->render_minor = -1;
    }

    /* XXX: I'm currently ignoring the colormap. Is that bad? */
    if (XC_SURFACE_RENDER_HAS_CREATE_PICTURE (surface))
	surface->picture = XRenderCreatePicture (dpy, drawable,
						 visual
						 ? XRenderFindVisualFormat (dpy, visual)
						 : XRenderFindStandardFormat (dpy, render_format),
						 0, NULL);

    surface->icimage = NULL;

    return surface;
}

XcSurface *
XcSurfaceCreateForIcImage (IcImage	*image)
{
    XcSurface *surface;

    surface = malloc (sizeof (XcSurface));
    if (surface == NULL)
	return NULL;

    surface->type = XC_SURFACE_TYPE_ICIMAGE;
    surface->transform = XC_XTRANSFORM_IDENTITY;

    surface->dpy = NULL;
    surface->gc = 0;
    surface->drawable = 0;
    surface->visual = NULL;
    surface->render_major = -1;
    surface->render_minor = -1;

    surface->picture = 0;

    surface->icimage = image;

    return surface;
}

void
XcSurfaceDestroy (XcSurface	*surface)
{
    if (surface->picture)
	XRenderFreePicture (surface->dpy, surface->picture);
	
    if (surface->icimage)
	IcImageDestroy (surface->icimage);

    free(surface);
}

/* XXX: Need to resolve PixRegion vs. Region confusion before this can work
void
XcSurfaceSetClipRegion (XcSurface	*surface,
			Region		region)
{
    if (surface->picture)
	XRenderSetPictureClipRegion (surface->dpy, surface->picture, region);

    if (surface->icimage)
	IcImageSetClipRegion (surface->icimage, region);
}
*/

void
XcSurfaceSetAlphaMap (XcSurface	*surface,
		      XcSurface	*alpha)
{
    if (surface->picture) {
	XRenderPictureAttributes pa;
	pa.alpha_map = alpha->picture;
	XRenderChangePicture (surface->dpy, surface->picture, CPAlphaMap, &pa);
    }

    if (surface->icimage) {
	/* XXX: Is libic ready for external alpha? */
    }
}

void
XcSurfaceSetTransform (XcSurface	*surface,
		       XTransform	*transform)
{
    surface->transform = *transform;

    if (surface->picture) {
	if (XC_SURFACE_RENDER_HAS_PICTURE_TRANSFORM (surface))
	    XRenderSetPictureTransform (surface->dpy, surface->picture, transform);
	/* XXX: Need support here if using an old RENDER without support
           for SetPictureTransform */
    }

    /* XXX: This cast should only occur with a #define hint from libic that it is OK */
    if (surface->icimage) {
	IcImageSetTransform (surface->icimage, (IcTransform *) transform);
    }
}

void
XcSurfaceGetTransform (XcSurface	*surface,
		       XTransform	*transform)
{
    *transform = surface->transform;
}

void
XcSurfaceSetRepeat (XcSurface	*surface,
		    int		repeat)
{
    if (surface->picture) {
	unsigned long mask;
	XRenderPictureAttributes pa;
	
	mask = CPRepeat;
	pa.repeat = repeat;

	XRenderChangePicture (surface->dpy, surface->picture, mask, &pa);
    }

    if (surface->icimage) {
	IcImageSetRepeat (surface->icimage, repeat);
    }
}

/* XXX: The Render specification has capitalized versions of these
   strings. However, the current implementation is case-sensitive and
   expects lowercase versions. */
static char *
_XcRenderFilterName (XcFilter filter)
{
    switch (filter) {
    case XcFilterFast:
	return "fast";
    case XcFilterGood:
	return "good";
    case XcFilterBest:
	return "best";
    case XcFilterNearest:
	return "nearest";
    case XcFilterBilinear:
	return "bilinear";
    default:
	return "best";
    }
}

void
XcSurfaceSetFilter (XcSurface	*surface,
		    XcFilter	filter)
{
    if (surface->picture) {
	XRenderSetPictureFilter (surface->dpy, surface->picture, _XcRenderFilterName (filter), NULL, 0);
    }

    if (surface->icimage) {
	IcImageSetFilter (surface->icimage, filter);
    }
}

static void
_XcSurfaceEnsureGC (XcSurface *surface)
{
    if (surface->gc)
	return;

    surface->gc = XCreateGC (surface->dpy, surface->drawable, 0, NULL);
}

void
XcSurfacePutImage (XcSurface		*surface,
		   unsigned char	*data,
		   unsigned int		width,
		   unsigned int		height,
		   unsigned int		stride)
{
    if (surface->picture) {
	XImage *image;
	unsigned bitmap_pad;

	/* XXX: This is obviously bogus. depth needs to be figured out for real */
	int depth = 32;

	if (depth > 16)
	    bitmap_pad = 32;
	else if (depth > 8)
	    bitmap_pad = 16;
	else
	    bitmap_pad = 8;

	image = XCreateImage(surface->dpy,
			     DefaultVisual(surface->dpy, DefaultScreen(surface->dpy)),
			     depth, ZPixmap, 0,
			     data, width, height,
			     bitmap_pad,
			     stride);
	/* XXX: Do we want to introduce a better error-mechanism for Xc? */
	if (image == NULL)
	    return;

	_XcSurfaceEnsureGC (surface);
	XPutImage(surface->dpy, surface->drawable, surface->gc,
		  image, 0, 0, 0, 0, width, height);

	/* Foolish XDestroyImage thinks it can free my data, but I won't
	   stand for it. */
	image->data = NULL;
	XDestroyImage(image);
    } else {
	/* XXX: Need to implement the IcImage method of setting a picture. memcpy? */
    }
}

void
_XcSurfacePullImage (XcSurface *surface)
{
/* XXX: NYI
    if (surface->type == XC_SURFACE_TYPE_ICIMAGE)
	return;

    if (surface->icimage) {
	IcImageDestroy (surface->icimage);
	surface->icimage = NULL;
    }

    _XcSurfaceEnsureGC (surface);
    surface->ximage = XGetImage (surface->dpy,
				 surface->drawable,
				 surface->gc,
				 0, 0,
				 width, height,
				 AllPlanes, ZPixmap);
    
    surface->icimage = IcImageCreateForData (image->data,
					     IcFormat *format,
					     int width, int height,
					     int bpp, int stride);
*/

    /* XXX: Add support here for pictures with external alpha. */

}

void
_XcSurfacePushImage (XcSurface *surface)
{
/* XXX: NYI
    if (surface->type == XC_SURFACE_TYPE_ICIMAGE)
	return;

    if (surface->ximage == NULL)
	return;

    _XcSurfaceEnsureGC (XcSurface *surface);
    XPutImage (surface->dpy,
	       surface->drawable,
	       surface->gc,
	       surface->ximage,
	       0, 0,
	       0, 0,
	       width, height);

    * Foolish XDestroyImage thinks it can free my data, but I won't
       stand for it. *
    surface->ximage->data = NULL;
    XDestroyImage(surface->ximage);
    surface->ximage = NULL;
*/
}

